-- set all your epic branding here
local configname = "SgConfig.cfg"
local firstrunmsg = "Sunflower's Garden configuration menu"

-- set your hardcoded priorities here
local menupriorities = {
	HUD = 100,
	Vanilla = 99,
	Other = 98,
	Nametags = 97,
	Hitfeed = 96,
	Hostmod = 95,
	["Mini Rankings"] = 94,
	["Map Records"] = 93,
	Dynmusic = 92,
	Sprays = 91,
	Elimination = 90,
}

rawset(_G, "CONFIG_Queue", CONFIG_Queue or {})

-- lots of "global" variables
local configmenus = {}
local config = false -- opened?
local loadconfig = true
local firstrun = true
local savechanges = false
local curmenu = 1
local curselection = 0
local presstics = 0
local pressrepeat = false
local curitems
local input = { [0] = {}, [1] = {}, [2] = {}, [3] = {} }
local hudmoving = false
local sitem -- selected item

-- this executes locally; show the config menu for the player, not any spectators!
-- as a side effect, we can't freeze the player by setting pw_nocontrol
-- workaround: don't use any buttons for controlling the menu (except drift)
COM_AddCommand("config", function(p)
	-- you cannot get inputs from the replay viewer
	-- ...unless reading their freecam movement counts as getting input
	if replayplayback then
		print("Sorry, this can't be used during replays.")
		return
	end

	local nobody = true
	for plr in players.iterate do
		nobody = false
	end
	if nobody then
		print("There's zero players here... are you at the title screen?")
		return
	end

	config = true
end, COM_LOCAL)

-- perform a long list of checks, then add a table of cvars to the menus
-- returns true if adding item succeeded
local function AddItem(item)
	local name = item[1] or item.name
	--local defaultvalue = item[2] or item.defaultvalue
	--local flags = item[3] or item.flags
	local possiblevalue = item[4] or item.possiblevalue or item.PossibleValue
	--local func = item[5] or item.func

	local menuname = item.config_menu
	local displayname = item.displayname
	local description = item.description
	local multi = item.config_multi
	local hudmove = item.config_hudmove
	local hudcall = item.config_hudcall
	local global = item.config_global

	if type(name) ~= "string" then error("Cvar name must be a string!", 2) end
	if type(menuname) ~= "string" then error(name..": Menu name must be a string!") end
	if type(displayname) ~= "string" then error(name..": Display name must be a string!", 2) end
	if type(description) ~= "string" then error(name..": Description must be a string!", 2) end
	if possiblevalue and type(possiblevalue) ~= "table" and type(possiblevalue) ~= "userdata" then
		error(name..": possiblevalue must be CV_*, nil, or a table of values!", 2)
	end

	-- splitscreen cvars, contains table of cvar names for the other players (cam2_dist, cam3_dist...)
	if multi then
		if type(multi) ~= "table" then error(name..": multi must be a table!", 2) end
	end

	-- hud moving cvar, set to name of Y offset cvar
	if hudmove then
		if type(hudmove) ~= "string" and hudmove ~= true then error(name..": hudmove must be a string (name of Y offset cvar) or true!", 2) end
		if hudcall and type(hudcall) ~= "function" then error(name..": hudcall must be a function!", 2) end
	end

	-- swap keys and values for easier indexing
	local possiblekeys
	if type(possiblevalue) == "table" then
		possiblekeys = {}
		for k, v in pairs(possiblevalue) do
			if k == "MIN" or k == "MAX" then
				-- nope
				possiblevalue = nil
				break
			end
			possiblekeys[v] = tostring(k)
		end
	end

	-- finally, add it to the menu
	local var = {
		["name"] = name,
		["displayname"] = displayname,
		["description"] = description,
		["possiblevalue"] = possiblevalue,
		["possiblekeys"] = possiblekeys,
		["multi"] = multi,
		["hudmove"] = hudmove,
		["hudcall"] = hudcall,
		["global"] = global,
	}

	-- look for an existing menu
	for _, menu in pairs(configmenus) do
		if menu.name == menuname then
			table.insert(menu.items, var)
			return
		end
	end

	-- otherwise, make a new one
	local p = menupriorities[menuname] or 0
	table.insert(configmenus, { name = menuname, items = {var}, priority = p })
	table.sort(configmenus, function(a, b) return a.priority > b.priority end)
end


-- hoo boy...
-- to work around precision loss, we'll multiply the fractional part to a precision of 5 digits with an epsilon
-- then truncate it down to two digits. yeah, that's totally sane
-- but hey, at least it lets us easily remove trailing zero(es)!
local function ToFloat(value)
	if value & (FRACUNIT-1) then
		if value < 0 then
			-- in two's complement, there is no such thing as "negative zero"
			local sign = value/FRACUNIT == 0 and "-" or ""
			local frac = string.format("%05d", FixedMul(-value & (FRACUNIT-1), 100100))
			frac = string.sub(frac, 1, 2):gsub("(.)0", "%1") -- remove trailing zero
			return sign..value/FRACUNIT.."."..frac
		else
			local frac = string.format("%05d", FixedMul(value & (FRACUNIT-1), 100100))
			if tonumber(frac) > 100000 then value = $ + FRACUNIT end -- carry out
			frac = string.sub(frac, 1, 2):gsub("(.)0", "%1") -- remove trailing zero
			return value/FRACUNIT.."."..frac
		end
	else
		return tostring(value/FRACUNIT)
	end
end

-- check if all the cvars in a menu actually exist
-- removes any offending cvars
-- also checks wheter to use large or small increments for float cvars
-- also saves pointer to cvar
local function CheckMenuItems(menuitems)
	local ret = false
	repeat
		local removed = false
		for k, item in pairs(menuitems) do
			local cvar = CV_FindVar(item.name)
			if not cvar then
				print("\x82WARNING:\x80 cvar doesn't exist: "..item.name)
				table.remove(menuitems, k)
				removed, ret = true, true
				break
			end

			if cvar.flags & CV_NETVAR then
				print("\x82WARNING:\x80 is a netvar: "..item.name)
				table.remove(menuitems, k)
				removed, ret = true, true
				break
			end

			item.cvar = cvar
			if cvar.flags & CV_FLOAT then
				local oldval = cvar.value
				-- if the maximum value is 5 or less, add 0.01 instead of 1
				CV_Set(cvar, "6")
				if cvar.value <= 5*FRACUNIT then
					item.divider = 100
				end
				CV_Set(cvar, ToFloat(oldval))
			end
		end
	until removed == false
	return ret
end

-- some cvars use OnOff/YesNo, but may set the default value to an integer
-- some also (ab)use the case insensitivity
local function StringifyValues(item, value, default)
	if item.possiblevalue == CV_OnOff then
		value = tostring($ and "On" or "Off")
		default = (($ == "1") or ($:lower() == "on")) and "On" or "Off"
	elseif item.possiblevalue == CV_YesNo then
		value = tostring($ and "Yes" or "No")
		default = (($ == "1") or ($:lower() == "yes")) and "Yes" or "No"
	elseif type(item.possiblevalue) == "table" then
		value = tostring(item.possiblekeys[value])
		local defnum = tonumber(default)
		if defnum ~= nil then -- any number counts
			default = tostring(item.possiblekeys[defnum])
		end
	else
		value = tostring(value)
	end

	return value, default
end

local function SaveCvar(configfile, item, cvar)
	local value, default = cvar.value, cvar.defaultvalue
	if cvar.flags & CV_FLOAT then
		value = ToFloat($)
	else
		value, default = StringifyValues(item, $1, $2)
	end

	-- only write cvars that aren't at their default value
	-- ignore saved vars
	if value ~= default and cvar.flags & CV_SAVE == 0 then
		local ok, err = configfile:write(cvar.name.." "..cvar.value.."\n")
		if not ok then
			error("Can't write to "..configname..": "..err, 0)
		end
	end
end

-- save local configuration
local function SaveConfig()
	local configfile, err = io.open(configname, "w")
	if not configfile then
		error("Can't open "..configname..": "..err, 0)
	end
	for _, menu in pairs(configmenus) do
		CheckMenuItems(menu.items)
		for _, item in pairs(menu.items) do
			SaveCvar(configfile, item, item.cvar)
			if item.multi then
				for _, v in pairs(item.multi) do
					SaveCvar(configfile, item, CV_FindVar(v))
				end
			elseif type(item.hudmove) == "string" then
				SaveCvar(configfile, item, CV_FindVar(item.hudmove))
			end
		end
	end

	local ok, error = configfile:close()
	if not ok then
		error("Can't close "..configname..": "..err, 0)
	end
	print("\x83Saved "..configname.."!")
end

-- load local configuration
local function LoadConfig()
	local configfile, err, code = io.open(configname, "r")
	if not configfile then
		if code == 2 then return end -- don't print error if the file doesn't exist
		error("Can't open "..configname.."(code "..code.."): "..err, 0)
	end
	for line in configfile:lines() do
		local space = line:find(" ")
		if not space then
			error(configname.." is corrupted (no arguments)", 0)
		end
		local var = line:sub(1, space-1)
		local value = line:sub(space+1)
		if value:find(" ") then
			error(configname.." is corrupted (more than one argument)", 0)
		end
		local cvar = CV_FindVar(var)
		if not cvar then
			print(configname..": cvar "..var.." not found, ignoring")
			continue
		end
		CV_Set(cvar, value)
	end
	local ok, error = configfile:close()
	if not ok then
		error("Can't close "..configname..": "..err, 0)
	end
	print("\x83Loaded "..configname.."!")
end

local function AddValue(cvar, amount)
	if cvar.flags & CV_FLOAT then
		amount = $ * FRACUNIT
		amount = $ / (sitem.divider or 1)
		CV_Set(cvar, ToFloat(cvar.value + amount))
	else
		CV_AddValue(cvar, amount)
	end
end

local function GetInput(i, dp)
	-- in 1P, don't let the specced player control your menu
	if i == 0 then dp = consoleplayer end

	if encoremode then
		input[i].right = dp.cmd.driftturn >= 200 and $ + 1 or 0 -- 200 to make life easier for controller
		input[i].left = dp.cmd.driftturn <= -200 and $ + 1 or 0 -- players with high sensitivity
	else
		input[i].left = dp.cmd.driftturn >= 200 and $ + 1 or 0
		input[i].right = dp.cmd.driftturn <= -200 and $ + 1 or 0
	end

	input[i].up = (dp.spectator and dp.cmd.forwardmove > 0 or (dp.cmd.buttons & BT_FORWARD == BT_FORWARD)) and $ + 1 or 0
	input[i].down = (dp.spectator and dp.cmd.forwardmove < 0 or (dp.cmd.buttons & BT_BACKWARD == BT_BACKWARD)) and $ + 1 or 0
	if input[i].up and input[i].down then input[i].up, input[i].down = 0, 0 end

	input[i].default = dp.cmd.buttons & BT_DRIFT == BT_DRIFT and $ + 1 or 0
end

local hudtime = 0
local ingame = false

-- the good stuff
local function ConfigHud(v, p, c)
	-- check for queued items
	if CONFIG_Queue and next(CONFIG_Queue) then repeat
		AddItem(table.remove(CONFIG_Queue, 1))
	until not next(CONFIG_Queue) end

	if loadconfig then
		loadconfig = false
		LoadConfig()
	end

	-- only run for the ~~first~~ LAST player in splitscreen, because yay draw order!
	if not config or (c and c.pnum ~= (hudmoving and 1 or splitscreen + 1)) then return end

	if #configmenus == 0 then
		print("Nothing to see here...")
		config = false
		return
	end

	-- exit if any other buttons are pressed
	if consoleplayer.cmd.buttons & ~(BT_FORWARD | BT_BACKWARD | BT_DRIFT) then
		config = false
		hudmoving = false
		presstics = 0
		if savechanges then
			savechanges = false
			SaveConfig()
			S_StartSound(nil, sfx_s245)
		end
		return
	end

	hudtime = $ + 1
	local arrowofs = hudtime >> 2 & 1
	local cursorpatch = v.cachePatch("M_CURSOR")
	local highlight = CV_FindVar("menuhighlight").value or (G_BattleGametype() and V_REDMAP or V_SKYMAP)

	-- get input
	if ingame then
		for dp, i in displayplayers.iterate do GetInput(i, dp) end
	else
		GetInput(0)
	end

	-- before going any further, check if this is the first run this session
	if firstrun then
		v.fadeScreen(31, 5)
		v.drawString(160, 64, firstrunmsg, 0, "center")
		v.drawString(32, 88, "Aim forward/backward:", 0, "left")
		v.drawString(288, 88, "Select item", V_YELLOWMAP, "right")
		v.drawString(32, 96, "Turn left/right:", 0, "left")
		v.drawString(288, 96, "Change value", V_YELLOWMAP, "right")
		v.drawString(32, 104, "Drift:", 0, "left")
		v.drawString(288, 104, "Reset to default", V_YELLOWMAP, "right")
		v.drawString(160, 120, "Press Drift to continue.", highlight, "center")
		v.drawString(160, 128, "Press any other button to exit.", highlight, "center")

		if input[0].default then
			firstrun = false
			curitems = configmenus[curmenu].items
			CheckMenuItems(curitems)
			S_StartSound(nil, sfx_s235)
		end
		return
	end

	-- hud movement in progress?
	if not ingame then hudmoving = false end
	if hudmoving then
		local xvar = sitem.cvar
		local yvar = type(sitem.hudmove) == "string" and CV_FindVar(sitem.hudmove)

		if yvar then
			if input[0].up == 1 or input[0].up >= TICRATE/2 then
				AddValue(yvar, -1)
				S_StartSound(nil, sfx_menu1)
			elseif input[0].down == 1 or input[0].down >= TICRATE/2 then
				AddValue(yvar, 1)
				S_StartSound(nil, sfx_menu1)
			end
		end

		if input[0].left == 1 or input[0].left >= TICRATE/2 then
			AddValue(xvar, -1)
			S_StartSound(nil, sfx_menu1)
		elseif input[0].right == 1 or input[0].right >= TICRATE/2 then
			AddValue(xvar, 1)
			S_StartSound(nil, sfx_menu1)
		end

		-- callback for previewing changes
		if sitem.hudcall then sitem.hudcall(v, p, c, xvar, yvar) end

		-- draw current coordinates
		if yvar then
			v.drawFill(129, 95, 62, 11, 27)
			v.drawString(130, 96, "X "..xvar.value, 0, "thin")
			v.drawString(161, 96, "Y "..yvar.value, 0, "thin")
		else
			v.drawFill(144, 95, 32, 11, 27)
			v.drawString(145, 96, (xvar.flags & CV_FLOAT) and xvar.value / FRACUNIT or xvar.value, 0, "thin")
		end

		-- exiting
		hudmoving = input[0].default ~= 1
		return
	end

	-- prevent people from screwing up their variables
	-- if up/down is pressed, don't allow pressing left/right until all directions are released
	if input[0].up or input[0].down then pressrepeat = true end
	if not (input[0].up > 0 or input[0].down > 0 or input[0].left > 0 or input[0].right > 0) then pressrepeat = false end
	if pressrepeat then input[0].left, input[0].right = 0, 0 end

	-- move up/down
	if input[0].up == 1 or (curselection > 0 and input[0].up >= TICRATE/2)
	or input[0].down == 1 or (curselection < #curitems and input[0].down >= TICRATE/2) then
		if input[0].up then
			curselection = $ - 1
			if curselection < 0 then curselection = #curitems end
		elseif input[0].down then
			curselection = $ + 1
			if curselection > #curitems then curselection = 0 end
		end
		S_StartSound(nil, sfx_menu1)
		sitem = curitems[curselection]
	-- switch to a different menu
	elseif curselection == 0 and (input[0].left == 1 or input[0].right == 1) then
		if input[0].left then
			curmenu = $ - 1
			if curmenu == 0 then curmenu = #configmenus end
		elseif input[0].right then
			curmenu = $ + 1
			if curmenu == #configmenus + 1 then curmenu = 1 end
		end
		curitems = configmenus[curmenu].items
		CheckMenuItems(curitems)
		S_StartSound(nil, sfx_s240)
	end

	-- change a cvar's value
	-- CV_AddValue will take care of the variable boundaries for us
	if curselection > 0 then
		presstics = 0
		for dp, i in displayplayers.iterate do
			local cvar = sitem.cvar
			if i > 0 then -- splitscreen cvar?
				if not (sitem.multi and ingame) then break end
				cvar = CV_FindVar(sitem.multi[i])
			end

			if (input[i].left == 1 or input[i].left >= TICRATE/2 or input[i].right == 1 or input[i].right >= TICRATE/2) then
				local change = 1
				if input[i].left then change = -$ end
				AddValue(cvar, change)
				S_StartSound(nil, sfx_menu1)
				savechanges = true

				if sitem.hudmove and ingame then
					hudmoving = true
					return
				end
			-- reset to default?
			elseif input[i].default == 1 then
				CV_Set(cvar, cvar.defaultvalue)
				if type(sitem.hudmove) == "string" then
					local yvar = CV_FindVar(sitem.hudmove)
					CV_Set(yvar, yvar.defaultvalue)
				end
				S_StartSound(nil, sfx_s229)
				savechanges = true
			end
		end
	end

	-- ok, time to actually draw shit
	local basey = 20

	-- fade the screen
	v.fadeScreen(31, 5)

	if presstics > 0 then v.draw(0, 0, v.cachePatch("k_itfree"), V_90TRANS - V_10TRANS*min((3*TICRATE/2) - presstics, 3)); presstics = $ - 1 end

	-- draw menu header
	local menuheader = configmenus[curmenu].name
	v.drawString(32, basey, menuheader, V_YELLOWMAP)
	if curselection == 0 then
		if input[0].default == TICRATE*2 then presstics = 3*TICRATE/2; S_StartSound(nil, sfx_itfree) end
		v.drawString(22 - arrowofs, basey, "\x1c", V_YELLOWMAP) -- left arrow
		v.drawString(34 + v.stringWidth(menuheader) + arrowofs, basey, "\x1d", V_YELLOWMAP) -- right arrow
	end

	-- find the longest menu option, use that as X offset for values
	local valofs = 0
	for _, item in pairs(curitems) do
		valofs = max($, v.stringWidth(item.displayname))
	end
	valofs = $ + 16

	-- now draw the list
	for k, item in ipairs(curitems) do
		local selected = curselection == k
		local value = item.cvar.value
		local default = item.cvar.defaultvalue
		local disabled = item.global and not CV_FindVar(item.global).value and V_GRAYMAP or 0
		local y = basey + k*8

		if item.cvar.flags & CV_FLOAT then
			value = ToFloat($)
		else
			value, default = StringifyValues(item, $1, $2)
			if item.hudmove then
				local yvar = CV_FindVar(item.hudmove)
				value = $..", "..yvar.value
				default = $..", "..yvar.defaultvalue
			end
		end

		-- draw display name
		v.drawString(32, y, item.displayname, disabled or selected and highlight or 0)

		-- draw cursor
		if selected then v.draw(8, y, cursorpatch) end

		-- draw value
		if not (item.multi and selected and splitscreen > 0 and ingame) then
			v.drawString(32 + valofs, y, value, (value ~= default and V_YELLOWMAP or disabled) | V_ALLOWLOWERCASE)
			if selected then -- arrows
				v.drawString(22 + valofs - arrowofs, y, "\x1c", highlight) -- left arrow
				v.drawString(34 + valofs + v.stringWidth(value) + arrowofs, y, "\x1d", highlight) -- right arrow
			end
		end
	end

	-- draw multi box
	if curselection > 0 and sitem.multi and splitscreen > 0 and ingame then
		local y = basey + curselection*8
		local longestval = 0

		-- once to get the longest value...
		for dp, i in displayplayers.iterate do
			local cvar = i == 0 and sitem.cvar or CV_FindVar(sitem.multi[i])
			local value = cvar.value
			if cvar.flags & CV_FLOAT then
				value = ToFloat($)
			else
				value = StringifyValues(sitem, $, cvar.defaultvalue)
			end
			longestval = max($, v.stringWidth(value))
		end

		-- ...to draw the box...
		v.drawFill(30 + valofs, 18 + curselection*8, 4+longestval, 12 + splitscreen*8, 27)

		-- ...and again to draw everyone's values
		for dp, i in displayplayers.iterate do
			local cvar = i == 0 and sitem.cvar or CV_FindVar(sitem.multi[i])
			local value = cvar.value
			if cvar.flags & CV_FLOAT then
				value = ToFloat($)
			else
				value = StringifyValues(sitem, $, cvar.defaultvalue)
			end
			v.drawString(32 + valofs, y + i*8, SG_Color2Chat[dp.skincolor]..value)
			v.drawString(22 + valofs - arrowofs, y + i*8, SG_Color2Chat[dp.skincolor].."\x1c") -- left arrow
			v.drawString(34 + valofs + v.stringWidth(value) + arrowofs, y + i*8, SG_Color2Chat[dp.skincolor].."\x1d") -- right arrow
		end
	end

	-- draw description at the bottom
	local name, descstr, disabled
	if curselection > 0 then
		name = sitem.name
		descstr = sitem.description
		disabled = sitem.global and not CV_FindVar(sitem.global).value
	else -- or if header is selected, draw a list of menus
		name = ""
		descstr = ""
		disabled = false
		for k, menu in ipairs(configmenus) do
			if curmenu == k then
				descstr = $.."\x82"..menu.name.."\x86 "
			else
				descstr = $.."\x86"..menu.name.." "
			end
		end
	end

	-- break it up into lines
	local description, numlines = SG_Linebreak(descstr, 38*8, v.stringWidth)

	-- then draw the description...
	v.drawString(8, 192 - numlines*8, description, V_ALLOWLOWERCASE|V_RETURN8|V_SNAPTOBOTTOM)
	-- ...and real name
	v.drawString(8, 184 - numlines*8, name, highlight|V_ALLOWLOWERCASE|V_SNAPTOBOTTOM)
	-- ...and status
	if disabled then v.drawString(312, 184 - numlines*8, "DISABLED", V_REDMAP|V_SNAPTOBOTTOM, "right") end
end

hud.add(function(v, p, c)
	ingame = true
	ConfigHud(v, p, c)
end, "game")

if FEATURE_INTERMISSIONHUD then
	hud.add(function(v)
		ingame = false
		ConfigHud(v)
	end, "intermission")
end

if FEATURE_VOTEHUD then
	hud.add(function(v)
		ingame = false
		ConfigHud(v)
	end, "vote")
end
